/* config.h inclusion marker */
#define CONFIG_H_WAS_INCLUDED

/* Whether system and steady clocks use the same duration type */
#define CHRONO_SAME_DURATION

/* Enable low-performing debugging facilities? */
#define ENABLE_DEBUG

/* Check logger messages? */
#define ENABLE_LOGGER_CHECKS

/* Fuzzing enabled. */
#undef FUZZING

/* AFL compiler enabled. */
#undef HAVE_AFL

/* Whether gtest defines edit_distance::CreateUnifiedDiff */
#define HAVE_CREATE_UNIFIED_DIFF

/* gss_str_to_oid is available */
#define HAVE_GSS_STR_TO_OID

/* MySQL is present */
#define HAVE_MYSQL

/* MySQL has mysql_get_option defined. */
#define HAVE_MYSQL_GET_OPTION

/* MySQL uses my_bool */
#define HAVE_MYSQL_MY_BOOL

/* NETCONF capabilities enabled */
/* #undef HAVE_NETCONF */

/* Check for optreset? */
#undef HAVE_OPTRESET

/* PostgreSQL is present */
#define HAVE_PGSQL

/* PostgreSQL connection parameter tcp_user_timeout supported */
#define HAVE_PGSQL_TCP_USER_TIMEOUT

/* Whether sockaddr has a sa_len member, and corresponding sin_len and sun_len */
#undef HAVE_SA_LEN

/* Whether you have the <sys/filio.h> header file. */
#undef HAVE_SYS_FILIO_H

/* Check valgrind headers */
/* #undef HAVE_VALGRIND_HEADERS */

/* Whether libc is musl */
#undef LIBC_MUSL

/* Explicit initialization of log4cplus possible */
#define LOG4CPLUS_INITIALIZER_H

/* Running on BSD? */
/* #undef OS_BSD */

/* Running on FreeBSD? */
/* #undef OS_FREEBSD */

/* Running on Linux? */
#define OS_LINUX

/* Running on NetBSD? */
/* #undef OS_NETBSD */

/* Running on OpenBSD? */
/* #undef OS_OPENBSD */

/* Running on OSX? */
/* #undef OS_OSX */

/* Running on Solaris? */
/* #undef OS_SOLARIS */

/* Name of package */
#define PACKAGE kea

/* Define to the address where bug reports for this package should be sent. */
/* #undef PACKAGE_BUGREPORT */

/* Define to the full name of this package. */
#define PACKAGE_NAME kea

/* Define to the full name and version of this package. */
/* #undef PACKAGE_STRING */

/* Define to the version of this package. */
#define PACKAGE_VERSION 3.0.3-git

/* Was Kea statically linked? */
/* #undef USE_STATIC_LINK */

/* Version number of package */
#define VERSION "3.0.3-git"

/* Compile with Botan crypto */
/* #undef WITH_BOTAN */

/* Heimdal GSS-API implementation */
/* #undef WITH_HEIMDAL */

/* Compile with OpenSSL crypto */
#define WITH_OPENSSL
